/*
 * Copyright (c) 2003, Enrique Jose Garcia Cota
 * All rights reserved.
 *
 * This file is part of antlraux. antlraux is distributed under 
 * the BSD license. See the license terms in file license.txt.
 *
 */

package antlraux.context.asts;

import antlraux.context.types.Type;
import antlraux.util.LexInfoAST;
import antlr.collections.AST;

/**
 * This class can be used to modelize the AST of an expression. It features an
 * expression type (a {@link antlraux.context.types.Type}), an RValue, LValue,
 * and value.
 * <p>
 * Warning: {@link TypeAST}, {@link ExpressionAST} and {@link ScopeAST} copy
 * first child and sibling with their only constructor. See their constructor's
 * comments for details.
 */
public class ExpressionAST extends TypeAST {
    /**
     * An expression with left value set to true can be put on the left sign of
     * an assignation; otherwise, they can't. An example of expression with
     * LValue is a variable:
     * 
     * <pre>
     * a = 3;
     * </pre>
     * 
     * An expression without LValue could be a literal or an addition:
     * 
     * <pre>
     *   &quot;Hello&quot;=&quot;world&quot;; // Incorrect; &quot;Hello&quot; does not have LValue
     *   a+b = c;         // Incorrect: a+b has no LValue
     * </pre>
     * 
     * Keep in mind that these are examples. Of course your languajes could
     * consider having literals with lvalues...
     */
    protected boolean LValue = true;

    /**
     * An expression with no Rvalue cannot be used on the right side of an
     * assignation, because they are not "readable". The typical example is an
     * invocation to a method that returns nothing:
     * 
     * <pre>
     *   void voidFunction() {...}
     *   
     *   ...
     *   
     *   int a = voidFunction(); // Incorrect; Rvalue required
     * </pre>
     * 
     * Of course this is an example. Some languajes might find useful setting
     * LValue to true in this kind of function calls (thus recuperating a void
     * value).
     */
    protected boolean RValue = true;

    /**
     * This is a calculated value. Some initializations must be calculated in
     * compile-time; for example static attribute initialization:
     * 
     * <pre>
     * class MyDumbClass {
     *     public int staticInt = 1 + 5 + 6;
     * }
     * </pre>
     * 
     * The value for literal "1" will be the integer 1, the value for literal
     * "5" will be the integer 5, and the value for "6" will the the integer 6.
     * This way, "1+5+6" can be evaluated in compile time (and set to 12).
     * <p>
     * Other use for this field is in iterpreters: The only way of calculating
     * values is having them dynamically stored in the ASTs.
     */
    protected Object EValue = null;

    public ExpressionAST() {
        super();
    }

    /**
     * This constructor copies the first child and next sibling of
     * <tt>other</tt> using the {@link #initialize(AST)} method
     */
    public ExpressionAST(AST other) {
        super();
        initialize(other);
    }

    public void setLValue(boolean b) {
        LValue = b;
    }

    public void setRValue(boolean b) {
        RValue = b;
    }

    public void setEValue(Object o) {
        EValue = o;
    }

    public boolean getLValue() {
        return LValue;
    }

    public boolean getRValue() {
        return RValue;
    }

    public Object getEValue() {
        return EValue;
    }

    /**
     * Warning: this implementation of initialize copies the first child and
     * next sibling of <tt>ast</tt>.
     */
    public void initialize(AST ast) {
        super.initialize(ast);
        setFirstChild(ast.getFirstChild());
        setNextSibling(ast.getNextSibling());

        if (ast instanceof ExpressionAST) {
            ExpressionAST aux = (ExpressionAST) ast;
            this.expType = aux.expType;
            this.RValue = aux.RValue;
            this.LValue = aux.LValue;
            this.EValue = aux.EValue;
        }
    }

    protected static boolean verboseStringConversion = false;

    /**
     * If set to true, this adds some additional information to the string
     * generated by {@link #toString()} (shows the AST's type, RValue, LValue
     * and EValue)
     */
    public static void setVerboseStringConversion(boolean b) {
        verboseStringConversion = b;
    }

    public String toString() {
        StringBuffer sb = new StringBuffer(super.toString());

        if (verboseStringConversion) {
            sb.append(" (exp: type=");
            sb.append(expType);
            sb.append(", R=");
            sb.append(RValue ? "1, L=" : "0, L=");
            sb.append(LValue ? "1, " : "0, ");
            sb.append(EValue);
            sb.append(")");
        }

        return sb.toString();
    }

}